@@ -1,4 +1,4 @@
-//! Regions Sort
+//! `regions_sort`
 //!
 //! Based on:
 //! Omar Obeya, Endrias Kahssay, Edward Fan, and Julian Shun.
@@ -19,10 +19,27 @@
 //!             4.2.2.2/b: If I is bigger than O, keep the remainder of I in the queue and continue
 //!             4.2.2.2/c: If O is bigger than I, keep the remainder of O in the queue and continue
 //!             4.2.2.3: Swap items in C heading to O, with items in I destined for C (items in C may or may not be destined for O ultimately)
+//!
+//! ## Characteristics
+//!
+//!  * mostly in-place
+//!  * multi-threaded
+//!  * unstable
+//!
+//! ## Performance
+//!
+//! This typically performs worse than the other, simpler, multi-threaded algorithms such as
+//! `recombinating_sort` and `scanning_sort`, however it uses a very clever and efficient algorithm
+//! from a research paper that means for certain inputs and certain memory conditions it can provide
+//! the best performance due to minimizing work spent copying and moving things.
+//!
+//! ## Notes
+//!
+//! This may not be entirely the same as the algorithm described by the research paper. Some things
+//! did not seem to matter, and have been omitted for performance reasons.
 
-use crate::director::director;
+use crate::sorter::Sorter;
 use crate::sorts::ska_sort::ska_sort;
-use crate::tuner::Tuner;
 use crate::utils::*;
 use crate::RadixKey;
 use partition::partition_index;
@@ -49,7 +66,7 @@ struct Edge<'bucket, T> {
 fn generate_outbounds<'bucket, T>(
     bucket: &'bucket mut [T],
     local_counts: &[[usize; 256]],
-    global_counts: &[usize],
+    global_counts: &[usize; 256],
 ) -> Vec<Edge<'bucket, T>> {
     let mut outbounds: Vec<Edge<T>> = Vec::new();
     let mut rem_bucket = bucket;
@@ -186,38 +203,26 @@ fn list_operations<T>(
     (outbounds, operations)
 }
 
-pub fn regions_sort<T>(bucket: &mut [T], level: usize) -> Vec<usize>
-where
+pub fn regions_sort<T>(
+    bucket: &mut [T],
+    counts: &[usize; 256],
+    tile_counts: &[[usize; 256]],
+    tile_size: usize,
+    level: usize,
+) where
     T: RadixKey + Sized + Send + Copy + Sync,
 {
-    if bucket.len() == 0 {
-        return get_counts(bucket, level).to_vec();
-    }
-
     let threads = current_num_threads();
-    let bucket_len = bucket.len();
-    let chunk_size = (bucket_len / threads) + 1;
-    let local_counts: Vec<[usize; 256]> = bucket
-        .par_chunks_mut(chunk_size)
-        .map(|chunk| {
-            let counts = get_counts(chunk, level);
+    bucket
+        .par_chunks_mut(tile_size)
+        .zip(tile_counts.par_iter())
+        .for_each(|(chunk, counts)| {
             let plateaus = detect_plateaus(chunk, level);
-            let (mut prefix_sums, end_offsets) = apply_plateaus(chunk, &counts, &plateaus);
+            let (mut prefix_sums, end_offsets) = apply_plateaus(chunk, &*counts, &plateaus);
             ska_sort(chunk, &mut prefix_sums, &end_offsets, level);
+        });
 
-            counts
-        })
-        .collect();
-
-    let mut global_counts = vec![0usize; 256];
-
-    local_counts.iter().for_each(|counts| {
-        for (i, c) in counts.iter().enumerate() {
-            global_counts[i] += *c;
-        }
-    });
-
-    let mut outbounds = generate_outbounds(bucket, &local_counts, &global_counts);
+    let mut outbounds = generate_outbounds(bucket, tile_counts, counts);
     let mut operations = Vec::new();
 
     // This loop calculates and executes all operations that can be done in parallel, each pass.
@@ -254,44 +259,57 @@ where
             }
         }
     }
-
-    global_counts
 }
 
-pub fn regions_sort_adapter<T>(
-    tuner: &(dyn Tuner + Send + Sync),
-    in_place: bool,
-    bucket: &mut [T],
-    level: usize,
-) where
-    T: RadixKey + Sized + Send + Copy + Sync,
-{
-    if bucket.len() <= 1 {
-        return;
+impl<'a> Sorter<'a> {
+    pub(crate) fn regions_sort_adapter<T>(
+        &self,
+        bucket: &mut [T],
+        counts: &[usize; 256],
+        tile_counts: &[[usize; 256]],
+        tile_size: usize,
+        level: usize,
+    ) where
+        T: RadixKey + Sized + Send + Copy + Sync,
+    {
+        if bucket.len() <= 1 {
+            return;
+        }
+
+        regions_sort(bucket, counts, tile_counts, tile_size, level);
+
+        if level == 0 {
+            return;
+        }
+
+        self.director(bucket, counts, level - 1);
     }
-
-    let global_counts = regions_sort(bucket, level);
-
-    if level == 0 {
-        return;
-    }
-
-    director(tuner, in_place, bucket, global_counts, level - 1);
 }
 
 #[cfg(test)]
 mod tests {
-    use crate::sorts::regions_sort::regions_sort_adapter;
-    use crate::test_utils::{sort_comparison_suite, NumericTest};
-    use crate::tuner::DefaultTuner;
+    use crate::sorter::Sorter;
+    use crate::tuners::StandardTuner;
+    use crate::utils::test_utils::{sort_comparison_suite, NumericTest};
+    use crate::utils::{aggregate_tile_counts, cdiv, get_tile_counts};
+    use rayon::current_num_threads;
 
     fn test_regions_sort<T>(shift: T)
     where
         T: NumericTest<T>,
     {
-        let tuner = DefaultTuner {};
+        let sorter = Sorter::new(true, &StandardTuner);
+
         sort_comparison_suite(shift, |inputs| {
-            regions_sort_adapter(&tuner, true, inputs, T::LEVELS - 1)
+            if inputs.len() == 0 {
+                return;
+            }
+
+            let tile_size = cdiv(inputs.len(), current_num_threads());
+            let tile_counts = get_tile_counts(inputs, tile_size, T::LEVELS - 1);
+            let counts = aggregate_tile_counts(&tile_counts);
+
+            sorter.regions_sort_adapter(inputs, &counts, &tile_counts, tile_size, T::LEVELS - 1);
         });
     }
 
